@@ -1,4 +1,4 @@
-from collections import OrderedDict, deque, ChainMap
+from collections import OrderedDict, deque, ChainMap, Iterable
 import uuid
 from .utils import (normalize_subs_input, root_ancestor,
                     separate_devices,
@@ -264,12 +264,11 @@ def print_summary_wrapper(plan):
             print('{motor.name} -> {args[0]}'.format(motor=msg.obj,
                                                      args=msg.args))
         elif cmd == 'create':
-            pass
+            read_cache = []
         elif cmd == 'read':
             read_cache.append(msg.obj.name)
         elif cmd == 'save':
             print('  Read {}'.format(read_cache))
-            read_cache = []
         yield msg
 
 
@@ -352,6 +351,42 @@ def subs_wrapper(plan, subs):
                                         _unsubscribe()))
 
 
+def suspend_wrapper(plan, suspenders):
+    """
+    Install suspenders to the RunEngine, and remove them at the end.
+
+    Parameters
+    ----------
+    plan : iterable or iterator
+        a generator, list, or similar containing `Msg` objects
+    suspenders : suspender or list of suspenders
+        Suspenders to use for the duration of the wrapper
+
+    Yields
+    ------
+    msg : Msg
+        messages from plan, with 'install_suspender' and 'remove_suspender'
+        messages inserted and appended
+    """
+    if not isinstance(suspenders, Iterable):
+        suspenders = [suspenders]
+
+    def _install():
+        for susp in suspenders:
+            yield Msg('install_suspender', None, susp)
+
+    def _remove():
+        for susp in suspenders:
+            yield Msg('remove_suspender', None, susp)
+
+    def _inner_plan():
+        yield from _install()
+        return (yield from plan)
+
+    return (yield from finalize_wrapper(_inner_plan(),
+                                        _remove()))
+
+
 def configure_count_time_wrapper(plan, time):
     """
     Preprocessor that sets all devices with a `count_time` to the same time.
@@ -913,6 +948,8 @@ def __read_and_stash_a_motor(obj, initial_positions, coupled_parents):
         reading = yield Msg('read', obj)
         if reading is None:
             # this plan may be being list-ified
+            print("*** all positions for {m.name} are "
+                  "relative to current position ***".format(m=obj))
             cur_pos = 0
         else:
             fields = getattr(obj, 'hints', {}).get('fields', [])
@@ -1016,19 +1053,6 @@ def reset_positions_wrapper(plan, devices=None):
     else:
         coupled_parents = set()
 
-    def read_and_stash_a_motor(obj):
-        try:
-            cur_pos = obj.position
-        except AttributeError:
-            reading = yield Msg('read', obj)
-            if reading is None:
-                # this plan may be being list-ified
-                cur_pos = 0
-            else:
-                k = list(reading.keys())[0]
-                cur_pos = reading[k]['value']
-        initial_positions[obj] = cur_pos
-
     def insert_reads(msg):
         eligible = devices is None or msg.obj in devices
         seen = msg.obj in initial_positions
@@ -1097,6 +1121,7 @@ def baseline_wrapper(plan, devices, name='baseline'):
 # Make generator function decorator for each generator instance wrapper.
 baseline_decorator = make_decorator(baseline_wrapper)
 subs_decorator = make_decorator(subs_wrapper)
+suspend_decorator = make_decorator(suspend_wrapper)
 relative_set_decorator = make_decorator(relative_set_wrapper)
 reset_positions_decorator = make_decorator(reset_positions_wrapper)
 # finalize_decorator is custom-made since it takes a plan as its
