@@ -7,11 +7,10 @@ from warnings import warn
 from inspect import Parameter, Signature
 from itertools import count, tee
 from collections import deque, defaultdict, ChainMap
-import signal
 from enum import Enum
 import functools
 import inspect
-
+from contextlib import ExitStack
 
 import jsonschema
 from event_model import DocumentNames, schemas
@@ -19,13 +18,15 @@ from super_state_machine.machines import StateMachine
 from super_state_machine.extras import PropertyMachine
 from super_state_machine.errors import TransitionError
 
-from .utils import (CallbackRegistry, SignalHandler, normalize_subs_input,
+from .utils import (CallbackRegistry, SigintHandler, normalize_subs_input,
                     AsyncInput, new_uid, NoReplayAllowed,
                     RequestAbort, RequestStop, RunEngineInterrupted,
                     IllegalMessageSequence, FailedPause, FailedStatus,
                     InvalidCommand, PlanHalt, Msg, ensure_generator,
                     single_gen, short_uid)
 
+_validate = functools.partial(jsonschema.validate, types={'array': (list, tuple)})
+
 
 class RunEngineStateMachine(StateMachine):
     """
@@ -107,13 +108,22 @@ class RunEngine:
     loop : asyncio event loop
         e.g., ``asyncio.get_event_loop()`` or ``asyncio.new_event_loop()``
 
-    preprocessors : list
+    preprocessors : list, optional
         Generator functions that take in a plan (generator instance) and
         modify its messages on the way out. Suitable examples include
         the functions in the module ``bluesky.plans`` with names ending in
         'wrapper'.  Functions are composed in order: the preprocessors
         ``[f, g]`` are applied like ``f(g(plan))``.
 
+    context_managers : list, optional
+        Context managers that will be entered when we run a plan. The context
+        managers will be composed in order, much like the preprocessors. If
+        this argument is omitted, we will use a user-oriented handler for
+        SIGINT. The elements of this list will be passed this ``RunEngine``
+        instance as their only argument. You may pass an empty list if you
+        would like a ``RunEngine`` with no signal handling and no context
+        managers.
+
     md_validator : callable, optional
         a function that raises and prevents starting a run if it deems
         the metadata to be invalid or incomplete
@@ -181,15 +191,19 @@ class RunEngine:
         It is set to ``bluesky.run_engine.PAUSE_MSG`` by default and
         can be modified based on needs.
 
+    commands:
+        The list of commands available to Msg.
+
     """
 
     state = LoggingPropertyMachine(RunEngineStateMachine)
     _UNCACHEABLE_COMMANDS = ['pause', 'subscribe', 'unsubscribe', 'stage',
                              'unstage', 'monitor', 'unmonitor', 'open_run',
-                             'close_run']
+                             'close_run', 'install_suspender',
+                             'remove_suspender']
 
     def __init__(self, md=None, *, loop=None, preprocessors=None,
-                 md_validator=None):
+                 context_managers=None, md_validator=None):
         if loop is None:
             loop = asyncio.get_event_loop()
         self._loop = loop
@@ -207,6 +221,9 @@ class RunEngine:
         if preprocessors is None:
             preprocessors = []
         self.preprocessors = preprocessors
+        if context_managers is None:
+            context_managers = [SigintHandler]
+        self.context_managers = context_managers
         if md_validator is None:
             md_validator = _default_md_validator
         self.md_validator = md_validator
@@ -224,13 +241,11 @@ class RunEngine:
         self._bundling = False  # if we are in the middle of bundling readings
         self._bundle_name = None  # name given to event descriptor
         self._deferred_pause_requested = False  # pause at next 'checkpoint'
-        self._sigint_handler = None  # intercepts Ctrl+C
-        self._last_sigint_time = None  # time most recent SIGINT was processed
-        self._num_sigints_processed = 0  # count SIGINTs processed
         self._exception = None  # stored and then raised in the _run loop
         self._interrupted = False  # True if paused, aborted, or failed
         self._objs_read = deque()  # objects read in one Event
         self._read_cache = deque()  # cache of obj.read() in one Event
+        self._asset_docs_cache = deque()  # cache of obj.collect_asset_docs()
         self._staged = set()  # objects staged, not yet unstaged
         self._objs_seen = set()  # all objects seen
         self._movable_objs_touched = set()  # objects we moved at any point
@@ -254,7 +269,7 @@ class RunEngine:
         self._msg_cache = deque()  # history of processed msgs for rewinding
         self._rewindable_flag = True  # if the RE is allowed to replay msgs
         self._plan_stack = deque()  # stack of generators to work off of
-        self._response_stack = deque([None])  # resps to send into the plans
+        self._response_stack = deque()  # resps to send into the plans
         self._exit_status = 'success'  # optimistic default
         self._reason = ''  # reason for abort
         self._task = None  # asyncio.Task associated with call to self._run
@@ -278,6 +293,7 @@ class RunEngine:
             'clear_checkpoint': self._clear_checkpoint,
             'rewindable': self._rewindable,
             'pause': self._pause,
+            'resume': self._resume,
             'collect': self._collect,
             'kickoff': self._kickoff,
             'complete': self._complete,
@@ -289,7 +305,9 @@ class RunEngine:
             'open_run': self._open_run,
             'close_run': self._close_run,
             'wait_for': self._wait_for,
-            'input': self._input, }
+            'input': self._input,
+            'install_suspender': self._install_suspender,
+            'remove_suspender': self._remove_suspender, }
 
         # public dispatcher for callbacks
         # The Dispatcher's public methods are exposed through the
@@ -303,7 +321,59 @@ class RunEngine:
         self._subscribe_lossless = self.dispatcher.subscribe
         self._unsubscribe_lossless = self.dispatcher.unsubscribe
 
-        self.loop.call_soon(self._check_for_signals)
+    @property
+    def commands(self):
+        '''
+        The list of commands available to Msg.
+
+        See Also
+        --------
+        :meth:`RunEngine.register_command`
+        :meth:`RunEngine.unregister_command`
+        :meth:`RunEngine.print_command_registry`
+
+        Examples
+        --------
+        >>> from bluesky import RunEngine
+        >>> RE = RunEngine()
+        >>> # to list commands
+        >>> RE.commands
+        '''
+        # return as a list, not lazy loader, no surprises...
+        return list(self._command_registry.keys())
+
+    def print_command_registry(self, verbose=False):
+        '''
+        This conveniently prints the command registry of available
+        commands.
+
+        Parameters
+        ----------
+        Verbose : bool, optional
+        verbose print. Default is False
+
+        See Also
+        --------
+        :meth:`RunEngine.register_command`
+        :meth:`RunEngine.unregister_command`
+        :attr:`RunEngine.commands`
+
+        Examples
+        --------
+        >>> from bluesky import RunEngine
+        >>> RE = RunEngine()
+        >>> # Print a very verbose list of currently registered commands
+        >>> RE.print_command_registry(verbose=True)
+        '''
+        commands = "List of available commands\n"
+
+        for command, func in self._command_registry.items():
+            docstring = func.__doc__
+            if not verbose:
+                docstring = docstring.split("\n")[0]
+            commands = commands + "{} : {}\n".format(command, docstring)
+
+        return commands
 
     def subscribe(self, func, name='all'):
         """
@@ -392,6 +462,7 @@ class RunEngine:
         self._bundling = False
         self._objs_read.clear()
         self._read_cache.clear()
+        self._asset_docs_cache.clear()
         self._uncollected.clear()
         self._describe_cache.clear()
         self._config_desc_cache.clear()
@@ -414,7 +485,7 @@ class RunEngine:
         self._deferred_pause_requested = False
         self._plan_stack = deque()
         self._msg_cache = deque()
-        self._response_stack = deque([None])
+        self._response_stack = deque()
         self._exception = None
         self._run_start_uids.clear()
         self._exit_status = 'success'
@@ -424,8 +495,6 @@ class RunEngine:
         self._pardon_failures = asyncio.Event(loop=self.loop)
         self._plan = None
         self._interrupted = False
-        self._last_sigint_time = None
-        self._num_sigints_processed = 0
 
         # Unsubscribe for per-run callbacks.
         for cid in self._temp_callback_ids:
@@ -470,6 +539,8 @@ class RunEngine:
         See Also
         --------
         :meth:`RunEngine.unregister_command`
+        :meth:`RunEngine.print_command_registry`
+        :attr:`RunEngine.commands`
         """
         self._command_registry[name] = func
 
@@ -484,6 +555,8 @@ class RunEngine:
         See Also
         --------
         :meth:`RunEngine.register_command`
+        :meth:`RunEngine.print_command_registry`
+        :attr:`RunEngine.commands`
         """
         del self._command_registry[name]
 
@@ -558,7 +631,7 @@ class RunEngine:
                        seq_num=next(self._interruptions_counter),
                        data={'interruption': content},
                        timestamps={'interruption': ttime.time()})
-            jsonschema.validate(doc, schemas[DocumentNames.event])
+            _validate(doc, schemas[DocumentNames.event])
             self.dispatcher.process(DocumentNames.event, doc)
 
     def __call__(self, *args, **metadata_kw):
@@ -651,8 +724,10 @@ class RunEngine:
             self._plan_stack.append(single_gen(Msg('wait_for', None, futs)))
             self._response_stack.append(None)
 
-        # Intercept ^C.
-        with SignalHandler(signal.SIGINT, self.log) as self._sigint_handler:
+        # Handle all context managers
+        with ExitStack() as stack:
+            for mgr in self.context_managers:
+                stack.enter_context(mgr(self))
             self._task = self.loop.create_task(self._run())
             try:
                 self.loop.run_forever()
@@ -730,10 +805,11 @@ class RunEngine:
     def _resume_event_loop(self):
         # may be called by 'resume' or 'abort'
         self.state = 'running'
-        self._last_sigint_time = None
-        self._num_sigints_processed = 0
-        # Intercept ^C
-        with SignalHandler(signal.SIGINT, self.log) as self._sigint_handler:
+
+        # Handle all context managers
+        with ExitStack() as stack:
+            for mgr in self.context_managers:
+                stack.enter_context(mgr(self))
             if self._task.done():
                 return
             try:
@@ -766,6 +842,18 @@ class RunEngine:
         self._suspenders.add(suspender)
         suspender.install(self)
 
+    @asyncio.coroutine
+    def _install_suspender(self, msg):
+        """
+        See :meth: `RunEngine.install_suspender`
+
+        Expected message object is:
+
+            Msg('install_suspender', None, suspender)
+        """
+        suspender = msg.args[0]
+        self.install_suspender(suspender)
+
     def remove_suspender(self, suspender):
         """
         Uninstall a suspender.
@@ -783,6 +871,18 @@ class RunEngine:
             suspender.remove()
         self._suspenders.discard(suspender)
 
+    @asyncio.coroutine
+    def _remove_suspender(self, msg):
+        """
+        See :meth: `RunEngine.remove_suspender`
+
+        Expected message object is:
+
+            Msg('remove_suspender', None, suspender)
+        """
+        suspender = msg.args[0]
+        self.remove_suspender(suspender)
+
     def clear_suspenders(self):
         """
         Uninstall all suspenders.
@@ -852,9 +952,11 @@ class RunEngine:
             # queue up the cached messages
             self._plan_stack.append(new_plan)
             self._response_stack.append(None)
+
             self._plan_stack.append(single_gen(
                 Msg('rewindable', None, self.rewindable)))
             self._response_stack.append(None)
+
             # if there is a post plan add it between the wait
             # and the cached messages
             if post_plan is not None:
@@ -862,9 +964,15 @@ class RunEngine:
                     post_plan = post_plan()
                 self._plan_stack.append(ensure_generator(post_plan))
                 self._response_stack.append(None)
+
+            # tell the devices they are ready to go again
+            self._plan_stack.append(single_gen(Msg('resume', None, )))
+            self._response_stack.append(None)
+
             # add the wait on the future to the stack
             self._plan_stack.append(single_gen(Msg('wait_for', None, [fut, ])))
             self._response_stack.append(None)
+
             # if there is a pre plan add on top of the wait
             if pre_plan is not None:
                 if callable(pre_plan):
@@ -970,10 +1078,19 @@ class RunEngine:
         """
         pending_cancel_exception = None
         self._reason = ''
+        # sentinel to decide if need to add to the response stack or not
+        sentinel = object()
         try:
             self.state = 'running'
             while True:
+                assert len(self._response_stack) == len(self._plan_stack)
+                # set resp to the sentinel so that if we fail in the sleep
+                # we do not add an extra response
+                resp = sentinel
                 try:
+                    # the new response to be added
+                    new_response = None
+
                     # This 'yield from' must be here to ensure that
                     # this coroutine breaks out of its current behavior
                     # before trying to get the next message from the
@@ -987,17 +1104,25 @@ class RunEngine:
                     # that any of the 'async' exceptions get thrown in the
                     # correct place
                     yield from asyncio.sleep(0, loop=self.loop)
+                    # always pop off a result, we are either sending it back in
+                    # or throwing an exception in, in either case the left hand
+                    # side of the yield in the plan will be moved past
+                    resp = self._response_stack.pop()
                     # The case where we have a stashed exception
-                    if self._exception is not None:
+                    if (self._exception is not None or
+                            isinstance(resp, Exception)):
                         # throw the exception at the current plan
                         try:
                             msg = self._plan_stack[-1].throw(
-                                self._exception)
+                                self._exception or resp)
                         except Exception as e:
                             # The current plan did not handle it,
                             # maybe the next plan (if any) would like
                             # to try
                             self._plan_stack.pop()
+                            # we have killed the current plan, do not give
+                            # it a new response
+                            resp = sentinel
                             if len(self._plan_stack):
                                 self._exception = e
                                 continue
@@ -1011,13 +1136,15 @@ class RunEngine:
                             self._exception = None
                     # The normal case of clean operation
                     else:
-                        resp = self._response_stack.pop()
                         try:
                             msg = self._plan_stack[-1].send(resp)
                         # We have exhausted the top generator
                         except StopIteration:
                             # pop the dead generator go back to the top
                             self._plan_stack.pop()
+                            # we have killed the current plan, do not give
+                            # it a new response
+                            resp = sentinel
                             if len(self._plan_stack):
                                 continue
                             # or reraise to get out of the infinite loop
@@ -1028,6 +1155,9 @@ class RunEngine:
                             # pop the dead plan, stash the exception and
                             # go to the top of the loop
                             self._plan_stack.pop()
+                            # we have killed the current plan, do not give
+                            # it a new response
+                            resp = sentinel
                             if len(self._plan_stack):
                                 self._exception = e
                                 continue
@@ -1056,7 +1186,7 @@ class RunEngine:
                     # to top of the loop
                     except KeyError:
                         # TODO make this smarter
-                        self._exception = InvalidCommand(msg.command)
+                        new_response = InvalidCommand(msg.command)
                         continue
 
                     # try to finally run the command the user asked for
@@ -1064,19 +1194,19 @@ class RunEngine:
                         # this is one of two places that 'async'
                         # exceptions (coming in via throw) can be
                         # raised
-                        response = yield from coro(msg)
+                        new_response = yield from coro(msg)
+
                     # special case `CancelledError` and let the outer
                     # exception block deal with it.
                     except asyncio.CancelledError:
                         raise
                     # any other exception, stash it and go to the top of loop
                     except Exception as e:
-                        self._exception = e
+                        new_response = e
                         continue
                     # normal use, if it runs cleanly, stash the response and
                     # go to the top of the loop
                     else:
-                        self._response_stack.append(response)
                         continue
 
                 except KeyboardInterrupt:
@@ -1098,6 +1228,12 @@ class RunEngine:
                     if self._exception is None:
                         self._exception = e
                     pending_cancel_exception = e
+                finally:
+                    # if we poped a response and did not pop a plan, we need
+                    # to put the new response back on the stack
+                    if resp is not sentinel:
+                        self._response_stack.append(new_response)
+
         except (StopIteration, RequestStop):
             self._exit_status = 'success'
             # TODO Is the sleep here necessary?
@@ -1166,43 +1302,6 @@ class RunEngine:
         if pending_cancel_exception is not None:
             raise pending_cancel_exception
 
-    def _check_for_signals(self):
-        # Check for pause requests from keyboard.
-        if self.state.is_running and (not self._interrupted):
-            count = self._sigint_handler.count
-            if count > self._num_sigints_processed:
-                self._num_sigints_processed = count
-                self.log.debug("RunEngine caught a new SIGINT")
-                self._last_sigint_time = ttime.time()
-
-                if count == 1:
-                    # Ctrl-C once -> request a deferred pause
-                    if not self._deferred_pause_requested:
-                        self.loop.call_soon(self.request_pause, True)
-                        print("A 'deferred pause' has been requested. The "
-                              "RunEngine will pause at the next checkpoint. "
-                              "To pause immediately, hit Ctrl+C again in the "
-                              "next 10 seconds.")
-                elif count > 1:
-                    # - Ctrl-C twice within 10 seconds -> hard pause
-                    self.log.debug("RunEngine detected two SIGINTs. "
-                                   "A hard pause will be requested.")
-                    self.loop.call_soon(self.request_pause, False)
-            else:
-                # No new SIGINTs to process.
-                if self._num_sigints_processed > 0:
-                    if ttime.time() - self._last_sigint_time > 10:
-                        self.log.debug("It has been 10 seconds since the "
-                                       "last SIGINT. Resetting SIGINT "
-                                       "handler.")
-                        # It's been 10 seconds since the last SIGINT. Reset.
-                        self._num_sigints_processed = 0
-                        self._sigint_handler.count = 0
-                        self._sigint_handler.interrupted = False
-                        self._last_sigint_time = None
-
-        self.loop.call_later(0.1, self._check_for_signals)
-
     @asyncio.coroutine
     def _wait_for(self, msg):
         """Instruct the RunEngine to wait until msg.obj has completed. Better
@@ -1269,7 +1368,7 @@ class RunEngine:
         # Emit an Event Descriptor for recording any interruptions as Events.
         if self.record_interruptions:
             self._interruptions_desc_uid = new_uid()
-            dk = {'dtype': 'string', 'shape': None, 'source': 'RunEngine'}
+            dk = {'dtype': 'string', 'shape': [], 'source': 'RunEngine'}
             interruptions_desc = dict(time=ttime.time(),
                                       uid=self._interruptions_desc_uid,
                                       name='interruptions',
@@ -1351,6 +1450,7 @@ class RunEngine:
                                          "bundle is closed with a 'save' or "
                                          'drop' "message.")
         self._read_cache.clear()
+        self._asset_docs_cache.clear()
         self._objs_read.clear()
         self._bundling = True
         self._bundle_name = None  # default
@@ -1397,8 +1497,15 @@ class RunEngine:
             # add this object to the cache of things we have read
             self._objs_read.append(obj)
 
-            # stash the results
+            # Stash the results, which will be emitted the next time _save is
+            # called --- or never emitted if _drop is called instead.
             self._read_cache.append(ret)
+            # Ask the object for any resource or datum documents is has cached
+            # and cache them as well. Likewise, these will be emitted if and
+            # when _save is called.
+            if hasattr(obj, 'collect_asset_docs'):
+                self._asset_docs_cache.extend(
+                    obj.collect_asset_docs(*msg.args, **msg.kwargs))
 
         return ret
 
@@ -1470,7 +1577,7 @@ class RunEngine:
             doc = dict(descriptor=descriptor_uid,
                        time=ttime.time(), data=data, timestamps=timestamps,
                        seq_num=next(seq_num_counter), uid=new_uid())
-            jsonschema.validate(doc, schemas[DocumentNames.event])
+            _validate(doc, schemas[DocumentNames.event])
             self.dispatcher.process(DocumentNames.event, doc)
 
         self._monitor_params[obj] = emit_event, kwargs
@@ -1522,8 +1629,19 @@ class RunEngine:
         # read in this Event grouping.
         objs_read = frozenset(self._objs_read)
 
-        # Event Descriptor
+        # Event Descriptor documents
         desc_key = self._bundle_name
+
+        # This is a separate check because it can be reset on resume.
+        seq_num_key = desc_key
+        if seq_num_key not in self._sequence_counters:
+            counter = count(1)
+            counter_copy1, counter_copy2 = tee(counter)
+            self._sequence_counters[seq_num_key] = counter_copy1
+            self._teed_sequence_counters[seq_num_key] = counter_copy2
+        self._bundling = False
+        self._bundle_name = None
+
         d_objs, doc = self._descriptors.get(desc_key, (None, None))
         if d_objs is not None and d_objs != objs_read:
             raise RuntimeError("Mismatched objects read, expected {!s}, "
@@ -1551,27 +1669,24 @@ class RunEngine:
             descriptor_uid = new_uid()
             doc = dict(run_start=self._run_start_uid, time=ttime.time(),
                        data_keys=data_keys, uid=descriptor_uid,
-                       configuration=config, name=self._bundle_name,
+                       configuration=config, name=desc_key,
                        hints=hints, object_keys=object_keys)
             yield from self.emit(DocumentNames.descriptor, doc)
             self.log.debug("Emitted Event Descriptor with name %r containing "
-                           "data keys %r (uid=%r)", self._bundle_name,
+                           "data keys %r (uid=%r)", desc_key,
                            data_keys.keys(), descriptor_uid)
             self._descriptors[desc_key] = (objs_read, doc)
 
         descriptor_uid = doc['uid']
 
-        # This is a separate check because it can be reset on resume.
-        seq_num_key = desc_key
-        if seq_num_key not in self._sequence_counters:
-            counter = count(1)
-            counter_copy1, counter_copy2 = tee(counter)
-            self._sequence_counters[seq_num_key] = counter_copy1
-            self._teed_sequence_counters[seq_num_key] = counter_copy2
-        self._bundling = False
-        self._bundle_name = None
+        # Resource and Datum documents
+        for name, doc in self._asset_docs_cache:
+            # Add a 'run_start' field to the resource document on its way out.
+            if name == 'resource':
+                doc['run_start'] = self._run_start_uid
+            yield from self.emit(DocumentNames(name), doc)
 
-        # Events
+        # Event documents
         seq_num = next(self._sequence_counters[seq_num_key])
         event_uid = new_uid()
         # Merge list of readings into single dict.
@@ -1580,7 +1695,8 @@ class RunEngine:
             readings[key]['value'] = readings[key]['value']
         data, timestamps = _rearrange_into_parallel_dicts(readings)
         # Mark all externally-stored data as not filled so that consumers
-        # know that the corresponding data are identifies, not deferenced data.
+        # know that the corresponding data are identifies, not dereferenced
+        # data.
         filled = {k: False
                   for k, v in
                   self._descriptors[desc_key][1]['data_keys'].items()
@@ -1726,6 +1842,14 @@ class RunEngine:
                                          "run is open.")
         self._uncollected.discard(obj)
 
+        if hasattr(obj, 'collect_asset_docs'):
+            # Resource and Datum documents
+            for name, doc in self.collect_asset_docs():
+                # Add a 'run_start' field to the resource document on its way out.
+                if name == 'resource':
+                    doc['run_start'] = self._run_start_uid
+                yield from self.emit(DocumentNames(name), doc)
+
         named_data_keys = obj.describe_collect()
         # e.g., {name_for_desc1: data_keys_for_desc1,
         #        name for_desc2: data_keys_for_desc2, ...}
@@ -1947,6 +2071,25 @@ class RunEngine:
         """
         self.request_pause(*msg.args, **msg.kwargs)
 
+    @asyncio.coroutine
+    def _resume(self, msg):
+        """Request the run engine to resume
+
+        Expected message object is:
+
+            Msg('resume', defer=False, name=None, callback=None)
+
+        See RunEngine.resume() docstring for explanation of the three
+        keyword arguments in the `Msg` signature
+        """
+        # Re-instate monitoring callbacks.
+        for obj, (cb, kwargs) in self._monitor_params.items():
+            obj.subscribe(cb, **kwargs)
+        # Notify Devices of the resume in case they want to clean up.
+        for obj in self._objs_seen:
+            if hasattr(obj, 'resume'):
+                obj.resume()
+
     @asyncio.coroutine
     def _checkpoint(self, msg):
         """Instruct the RunEngine to create a checkpoint so that we can rewind
@@ -2164,7 +2307,7 @@ class RunEngine:
     @asyncio.coroutine
     def emit(self, name, doc):
         "Process blocking callbacks and schedule non-blocking callbacks."
-        jsonschema.validate(doc, schemas[name])
+        _validate(doc, schemas[name])
         self.dispatcher.process(name, doc)
 
 
